import logging

from tradeflow.constants import Logger

logging.basicConfig(format=Logger.FORMAT, level=Logger.LEVEL)


def get_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)
    return logger



import ctypes as ct
import fnmatch
import os
from pathlib import Path
from typing import Literal, List

from statsmodels.tools.typing import ArrayLike1D

from tradeflow import logger_utils
from tradeflow.definitions import ROOT_DIR
from tradeflow.exceptions import TooManySharedLibrariesException

logger = logger_utils.get_logger(__name__)

ARGUMENT_TYPES = "argtypes"
RESULT_TYPES = "restype"

SHARED_LIBRARY_NAME = "libcpp"
SHARED_LIBRARY_EXTENSIONS = ["so", "dll", "dylib", "pyd"]

function_to_argtypes_and_restype = {
    "my_simulate": {
        # size (int), inverted_params (double*), constant_parameter (double), nb_params (int), last_signs (int*), seed (int), simulation (int*)
        ARGUMENT_TYPES: (ct.c_int, ct.POINTER(ct.c_double), ct.c_double, ct.c_int, ct.POINTER(ct.c_int), ct.c_int, ct.POINTER(ct.c_int)),
        RESULT_TYPES: ct.c_void_p
    }
}


def get_c_type(c_type_str: Literal["int", "double"]) -> ct._SimpleCData:
    """
    Return a ctypes type corresponding to a given C data type (in a string).

    Parameters:
    -----------
    c_type_str : Literal["int", "double"]
        A string indicating the desired C data type.

    Returns:
    --------
    ct._SimpleCData
        The corresponding ctypes type.
    """
    c_type_str_to_c_type = {
        "int": ct.c_int,
        "double": ct.c_double
    }

    if c_type_str not in c_type_str_to_c_type:
        raise Exception(f"Unknown type {c_type_str}")

    return c_type_str_to_c_type[c_type_str]


class CArray:

    @staticmethod
    def of(c_type_str: Literal["int", "double"], arr: ArrayLike1D) -> ct.Array:
        """
        Create a ctypes array from a Python list.

        Parameters
        ----------
        c_type_str : {'int', 'double'}
            The type of the array to be created.
        arr : array_like
            The array from which to create the ctypes array.

        Returns
        -------
        ct.Array
            The ctypes array containing the elements of `arr`.
        """
        c_type = get_c_type(c_type_str=c_type_str)
        return (c_type * len(arr))(*arr)


class CArrayEmpty:

    @staticmethod
    def of(c_type_str: Literal["int", "double"], size: int) -> ct.Array:
        """
        Create an empty ctypes array of a given size.

        Parameters
        ----------
        c_type_str : {'int', 'double'}
            The type of the array to be created.
        size : int
            The size of the ctypes array to create.

        Returns
        -------
        ct.Array
            The empty ctypes array of size `size`.
        """
        c_type_str = get_c_type(c_type_str=c_type_str)
        return (c_type_str * size)()


def load_shared_library() -> ct.CDLL:
    """
    Return the shared library of the project.

    Returns
    -------
    ct.CDLL
        The loaded shared library.
    """
    lib_file = get_shared_library_file(directory=ROOT_DIR, shared_library_name=SHARED_LIBRARY_NAME)
    shared_lib = ct.CDLL(lib_file, winmode=0)
    set_shared_library_functions(shared_lib=shared_lib)

    return shared_lib


def set_shared_library_functions(shared_lib: ct.CDLL) -> None:
    """
    Set argument and result types of functions in the shared library.

    Parameters
    ----------
    shared_lib : ct.CDLL
        The shared library for which to set argument and result types for all functions.
    """
    for function_name in function_to_argtypes_and_restype.keys():
        setattr(getattr(shared_lib, function_name), ARGUMENT_TYPES, function_to_argtypes_and_restype.get(function_name).get(ARGUMENT_TYPES))
        setattr(getattr(shared_lib, function_name), RESULT_TYPES, function_to_argtypes_and_restype.get(function_name).get(RESULT_TYPES))


def get_shared_library_file(directory: str, shared_library_name: str) -> str:
    """
    Return the path to the shared library `shared_library_name`.

    Parameters
    ----------
    directory : str
        The directory in which to search for the shared library.
    shared_library_name : str
        The name of the shared library.

    Returns
    -------
    str
        The path to the shared library, the extension of the file can be 'so' (Linux), 'dll' (Windows), 'dylib' (macOS), or 'pyd'.
    """
    shared_library_files = []
    for potential_extension in SHARED_LIBRARY_EXTENSIONS:
        shared_library_files.extend(find_files(pattern=f"{shared_library_name}.{potential_extension}", directory=directory))
        shared_library_files.extend(find_files(pattern=f"{shared_library_name}.*.{potential_extension}", directory=directory))

    if len(shared_library_files) == 0:
        raise FileNotFoundError(f"No shared library found for name '{shared_library_name}' with one of the extension in {SHARED_LIBRARY_EXTENSIONS} in directory {directory}.")
    if len(shared_library_files) >= 2:
        raise TooManySharedLibrariesException(f"{len(shared_library_files)} shared libraries found with name '{shared_library_name}' with extension in {SHARED_LIBRARY_EXTENSIONS} have been found: {', '.join(shared_library_files)} in directory: {directory}.")

    return str(Path(directory).joinpath(shared_library_files[0]))


def find_files(pattern: str, directory: str) -> List[str]:
    """
    Return files matching a specified pattern within a directory.

    Parameters
    ----------
    pattern : str
        The file name pattern to search for.
    directory : str
        The directory in which to search for files.

    Returns
    -------
    bool
        The file names matching the pattern (only the file names, not their full paths).
    """
    matched_files = []
    for root, dirs, files in os.walk(directory):
        if root == directory:
            for filename in fnmatch.filter(files, pattern):
                matched_files.append(filename)

    return matched_files



from __future__ import annotations

from typing import Literal, Optional

import numpy as np
from statsmodels.regression import yule_walker
from statsmodels.tools.typing import ArrayLike1D
from statsmodels.tsa.ar_model import ar_select_order, AutoReg

import tradeflow.ctypes_utils as ctypes_utils
from tradeflow import logger_utils
from tradeflow.constants import OrderSelectionMethodAR, FitMethodAR, InformationCriterion
from tradeflow.ctypes_utils import CArray, CArrayEmpty
from tradeflow.exceptions import IllegalValueException, ModelNotFittedException, IllegalNbLagsException, \
    NonStationaryTimeSeriesException
from tradeflow.general_utils import check_condition, check_enum_value_is_valid, get_enum_values, \
    is_value_within_interval_exclusive
from tradeflow.time_series import TimeSeries

logger = logger_utils.get_logger(__name__)


class AR(TimeSeries):
    """
    Autoregressive model for trade/order signs.

    Parameters
    ----------
    signs : array_like
        An array of signs where each element is either 1 (representing a buy) or -1 (representing a sell).
    max_order : int, default None
        The maximum order of the autoregressive model.
        If None, the maximum order is set to 12*(nobs/100)^{1/4} as outlined in Schwert (1989).
    order_selection_method : {'information_criterion', 'pacf'}, default None
        The method for selecting the order of the model. If None, the order of the model will be `max_order`.

        * 'information_criterion' - Choose the model that minimizes a given information criterion.
        * 'pacf' - Choose the model using the number of significant lags in the partial autocorrelation function of the time series of signs.

    information_criterion : {'aic', 'bic', 'hqic'}, optional
        The information criterion to use in the order selection.
        It has no effect if `order_selection_method` is not 'information_criterion'.
        * 'aic' - Akaike information criterion.
        * 'bic' - Bayesian information criterion.
        * 'hqic' - Hannan–Quinn information criterion.
    """

    def __init__(self, signs: ArrayLike1D, max_order: Optional[int] = None,
                 order_selection_method: Optional[Literal["information_criterion", "pacf"]] = None,
                 information_criterion: Optional[Literal["aic", "bic", "hqic"]] = None) -> None:
        super().__init__(signs=signs)
        self._max_order = self._init_max_order(max_order=max_order)
        self._order_selection_method = check_enum_value_is_valid(enum_obj=OrderSelectionMethodAR,
                                                                 value=order_selection_method,
                                                                 parameter_name="order_selection_method",
                                                                 is_none_valid=True)
        self._information_criterion = check_enum_value_is_valid(enum_obj=InformationCriterion,
                                                                value=information_criterion,
                                                                parameter_name="information_criterion",
                                                                is_none_valid=self._order_selection_method is None or self._order_selection_method == OrderSelectionMethodAR.PACF)
        if self._information_criterion is not None and (
                self._order_selection_method is None or self._order_selection_method == OrderSelectionMethodAR.PACF):
            logger.warning(
                f"The information criterion '{self._information_criterion}' will have no effect as the order selection method '{self._order_selection_method}' doesn't use it.")

        # Will be set during fit()
        self._constant_parameter = 0
        self._parameters = None

    def _init_max_order(self, max_order: Optional[int]) -> int:
        if max_order is None:
            # Schwert (1989)
            max_order = int(np.ceil(12.0 * np.power(len(self._signs) / 100.0, 1 / 4.0)))

        check_condition(condition=1 <= max_order < len(self._signs) // 2,
                        exception=IllegalNbLagsException(
                            f"{max_order} is not valid for 'max_order', it must be positive and lower than 50% of the time series length (< {len(self._signs) // 2})."))
        logger.info(f"The maximum order has been set to {max_order}.")
        return max_order

    def fit(self, method: Literal["yule_walker", "ols_with_cst"]) -> AR:
        """
        Estimate the model parameters.

        Parameters
        ----------
        method : {'yule_walker', 'ols_with_cst'}
            The method to use for estimating parameters.

            * 'yule_walker' - Use the Yule Walker equations to estimate model parameters.
              There will be no constant term, thus the percentage of buy signs
              in the time series generated with these parameters will be close to 50%.

            * 'ols_with_cst' - Use OLS to estimate model parameters.
              There will be a constant term, thus the percentage of buy signs in the time series
              generated with these parameters will be close to the one from the training time series.

        Returns
        -------
        AR
            The AR instance.
        """
        method = check_enum_value_is_valid(enum_obj=FitMethodAR, value=method, parameter_name="method",
                                           is_none_valid=False)
        self._select_order()

        if method == FitMethodAR.YULE_WALKER:
            check_condition(self._is_time_series_stationary(regression="n"), NonStationaryTimeSeriesException("The time series must be stationary to be fitted."))
            self._parameters = yule_walker(x=self._signs, order=self._order, method="mle", df=None, inv=False, demean=True)[0]
        elif method == FitMethodAR.OLS_WITH_CST:
            check_condition(self._is_time_series_stationary(regression="c"), NonStationaryTimeSeriesException("The time series must be stationary to be fitted."))
            ar_model = AutoReg(endog=self._signs, lags=self._order, trend="c").fit()
            self._constant_parameter, self._parameters = ar_model.params[0], ar_model.params[1:]
        else:
            raise IllegalValueException(
                f"The method '{method}' for the parameters estimation is not valid, it must be among {get_enum_values(enum_obj=FitMethodAR)}.")

        logger.info(f"The AR({self._order}) model has been fitted with method '{method}'.")
        return self

    def _select_order(self) -> None:
        if self._order_selection_method is None:
            self._order = self._max_order

        else:
            if self._order_selection_method == OrderSelectionMethodAR.INFORMATION_CRITERION:
                model = ar_select_order(endog=self._signs, maxlag=self._max_order,
                                        ic=self._information_criterion.value, trend="n")
                self._order = len(model.ar_lags)
            elif self._order_selection_method == OrderSelectionMethodAR.PACF:
                pacf_coeffs, confidence_interval = self.calculate_pacf(nb_lags=self._max_order, alpha=0.05)

                pacf_coeffs = pacf_coeffs[1:]
                confidence_interval = confidence_interval[1:]

                lower_band = confidence_interval[:, 0] - pacf_coeffs
                upper_band = confidence_interval[:, 1] - pacf_coeffs

                order = 0
                for acf_coeff, value_lower_band, value_upper_band in zip(pacf_coeffs, lower_band, upper_band):
                    if is_value_within_interval_exclusive(value=acf_coeff, lower_bound=value_lower_band,
                                                          upper_bound=value_upper_band):
                        break
                    order += 1
                self._order = order
            else:
                raise IllegalValueException(
                    f"The method '{self._order_selection_method}' for the order selection is not valid, it must be among {get_enum_values(enum_obj=OrderSelectionMethodAR)}")

        logger.info(
            f"AR order selection: {self._order} lags (method: {self._order_selection_method}, information criterion: {self._information_criterion}, time series length: {len(self._signs)}).")

    def simulate(self, size: int, seed: Optional[int] = None) -> np.ndarray:
        """
        Simulate a time series of signs after the model has been fitted.

        Parameters
        ----------
        size : int
            The number of signs to simulate.
        seed : int, default None
            Seed used to initialize the pseudo-random number generator.
            If `seed` is ``None``, then a random seed is used.

        Returns
        -------
        np.ndarray
            The simulated signs (+1 for buy, -1 for sell).
        """
        check_condition(size > 0, IllegalValueException(
            f"The size '{size}' for the time series to be simulated is not valid, it must be greater than 0."))
        check_condition(self._parameters is not None, ModelNotFittedException(
            "The model has not yet been fitted. Fit the model first by calling 'fit()'."))

        inverted_params = CArray.of(c_type_str="double", arr=self._parameters[::-1])
        last_signs = CArray.of(c_type_str="int", arr=np.array(self._signs[-self._order:]).astype(int))
        self._simulation = CArrayEmpty.of(c_type_str="int", size=size)

        cpp_lib = ctypes_utils.load_shared_library()
        cpp_lib.my_simulate(size, inverted_params, self._constant_parameter, len(inverted_params), last_signs, seed, self._simulation)
        return self._simulation[:]



import enum
from typing import List, Any, Optional, Union

from tradeflow.exceptions import EnumValueException


def check_condition(condition: bool, exception: Exception) -> None:
    """
    Raise an exception if a condition is false, otherwise do nothing.

    Parameters
    ----------
    condition : bool
        The condition to check.
    exception : Exception
        The exception to raise if the condition is false.

    Raises
    ------
    Exception
        If the condition is false.
    """
    if not condition:
        raise exception


def get_enum_values(enum_obj: enum) -> List[Any]:
    """
    Return a list containing the values of an enum.

    Parameters
    ----------
    enum_obj : enum.EnumType
        The enum for which the values will be retrieved.

    Returns
    -------
    list of any
        The values of the enum.
    """
    return [enum_object.value for enum_object in enum_obj]


def check_enum_value_is_valid(enum_obj: enum, value: Optional[object], is_none_valid: bool, parameter_name: str) -> Optional[enum.Enum]:
    """
    Raise an EnumValueException if a value is not within an enum.

    Parameters
    ----------
    enum_obj : enum.EnumType
        The enum to check the value against.
    value : object or None
        The value to verify.
    is_none_valid : bool
        Flag indicating whether None is allowed.
    parameter_name : str
        Variable name for exceptions.

    Returns
    -------
    enum.Enum or None
        Enum associated to the passed value or None if the value is None and is_none_valid is true.

    Raises
    ------
    EnumValueException
        If the value does not exist withing the enum or if the value is None and `is_none_valid` is false.
    """
    exception_message = f"The value '{value}' for {parameter_name} is not valid, it must be among {get_enum_values(enum_obj=enum_obj)} or None if it is valid."
    if value is None:
        if is_none_valid:
            return value
        else:
            raise EnumValueException(exception_message)

    try:
        enum_obj = enum_obj(value)
    except ValueError:
        raise EnumValueException(exception_message)
    else:
        return enum_obj


def is_value_within_interval_exclusive(value: Union[int, float], lower_bound: Union[int, float], upper_bound: Union[int, float]) -> bool:
    """
    Return whether a value is strictly within an interval or not.

    Parameters
    ----------
    value : int or float
        The value to check.
    lower_bound : int or float
        The strict lower bound of the interval.
    upper_bound : int or float
        The strict upper bound of the interval

    Returns
    -------
    bool
        True if the value is strictly within the interval, false otherwise.
    """
    return lower_bound < value < upper_bound



from typing import List

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext

PROJECT = "tradeflow"
SHARED_LIBRARY_NAME = "libcpp"
CPP_MODULES = ["simulate"]


class CppExtension(Extension):
    pass


class new_build_ext(build_ext):
    extra_compile_args = {
        f"{PROJECT}.{SHARED_LIBRARY_NAME}": {
            "unix": ["-std=c++17"],
            "msvc": ["/std:c++17"]
        }
    }

    def build_extension(self, ext):
        extra_args = self.extra_compile_args.get(ext.name)
        if extra_args is not None:
            compiler_type = self.compiler.compiler_type
            ext.extra_compile_args += extra_args.get(compiler_type, [])

        build_ext.build_extension(self, ext)

    def get_export_symbols(self, ext):
        return ext.export_symbols + list(map(lambda module: f"my_{module}", CPP_MODULES))


def build_cpp_extension() -> List[CppExtension]:
    cpp_extension = CppExtension(name=f"{PROJECT}.{SHARED_LIBRARY_NAME}",
                                 sources=[f"{PROJECT}/{cpp_module}.cpp" for cpp_module in CPP_MODULES],
                                 language="c++"
                                 )
    return [cpp_extension]


setup(
    packages=[PROJECT],
    ext_modules=build_cpp_extension(),
    cmdclass={'build_ext': new_build_ext}
)



import logging
from enum import Enum


class OrderSelectionMethodAR(Enum):
    INFORMATION_CRITERION = "information_criterion"
    PACF = "pacf"

    def __str__(self) -> str:
        return self._value_


class FitMethodAR(Enum):
    YULE_WALKER = "yule_walker"
    OLS_WITH_CST = "ols_with_cst"

    def __str__(self) -> str:
        return self._value_


class InformationCriterion(Enum):
    AIC = "aic"
    BIC = "bic"
    HQIC = "hqic"

    def __str__(self) -> str:
        return self._value_


class Logger:
    FORMAT = "%(asctime)s [%(filename)s] [%(levelname)s] - %(message)s"
    LEVEL = logging.INFO
